"""This module contains functions to manage models."""

from json import loads, dumps
from pathlib import Path
from xspect.models.probabilistic_filter_model import ProbabilisticFilterModel
from xspect.models.probabilistic_single_filter_model import (
    ProbabilisticSingleFilterModel,
)
from xspect.models.probabilistic_filter_svm_model import ProbabilisticFilterSVMModel
from xspect.definitions import get_xspect_model_path


def get_genus_model(genus):
    """Get a metagenomic model for the specified genus."""
    genus_model_path = get_xspect_model_path() / (genus.lower() + "-genus.json")
    genus_filter_model = ProbabilisticSingleFilterModel.load(genus_model_path)
    return genus_filter_model


def get_species_model(genus):
    """Get a species classification model for the specified genus."""
    species_model_path = get_xspect_model_path() / (genus.lower() + "-species.json")
    species_filter_model = ProbabilisticFilterSVMModel.load(species_model_path)
    return species_filter_model


def get_model_by_slug(model_slug: str):
    """Get a model by its slug."""
    model_path = get_xspect_model_path() / (model_slug + ".json")
    model_metadata = get_model_metadata(model_path)
    if model_metadata["model_class"] == "ProbabilisticSingleFilterModel":
        return ProbabilisticSingleFilterModel.load(model_path)
    elif model_metadata["model_class"] == "ProbabilisticFilterSVMModel":
        return ProbabilisticFilterSVMModel.load(model_path)
    elif model_metadata["model_class"] == "ProbabilisticFilterModel":
        return ProbabilisticFilterModel.load(model_path)
    else:
        raise ValueError(f"Model class {model_metadata['model_class']} not recognized.")


def get_model_metadata(model: str | Path):
    """Get the metadata of a model."""
    if isinstance(model, str):
        model_path = get_xspect_model_path() / (model + ".json")
    elif isinstance(model, Path):
        model_path = model
    else:
        raise ValueError("Model must be a string (slug) or a Path object.")

    if not model_path.exists() or not model_path.is_file():
        raise ValueError(f"Model at {model_path} does not exist.")

    with open(model_path, "r", encoding="utf-8") as file:
        model_json = loads(file.read())
        return model_json


def update_model_metadata(model_slug: str, author: str, author_email: str):
    """Update the metadata of a model."""
    model_metadata = get_model_metadata(model_slug)
    model_metadata["author"] = author
    model_metadata["author_email"] = author_email

    model_path = get_xspect_model_path() / (model_slug + ".json")
    with open(model_path, "w", encoding="utf-8") as file:
        file.write(dumps(model_metadata, indent=4))


def update_model_display_name(model_slug: str, filter_id: str, display_name: str):
    """Update the display name of a filter in a model."""
    model_metadata = get_model_metadata(model_slug)
    model_metadata["display_names"][filter_id] = display_name

    model_path = get_xspect_model_path() / (model_slug + ".json")
    with open(model_path, "w", encoding="utf-8") as file:
        file.write(dumps(model_metadata, indent=4))


def get_models():
    """Get a list of all available models in a dictionary by type."""
    model_dict = {}
    for model_file in get_xspect_model_path().glob("*.json"):
        model_metadata = get_model_metadata(model_file)
        model_type = model_metadata["model_type"]
        model_dict.setdefault(model_type, []).append(
            model_metadata["model_display_name"]
        )
    return model_dict



""" Module for defining the Pipeline class. """

import json
from pathlib import Path
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
from xspect.file_io import get_records_by_id
from xspect.models.result import StepType, SubprocessingStep
from xspect.run import Run
from xspect.models.result import ModelResult
from xspect.model_management import get_model_by_slug


class ModelExecution:
    """Class for storing a processing step of an XspecT pipeline."""

    def __init__(
        self,
        model_slug: str,
        sparse_sampling_step: int = 1,
    ):
        self.model_slug = model_slug
        self.sparse_sampling_step = sparse_sampling_step
        self.pipeline_steps = []

    def add_pipeline_step(
        self,
        pipeline_step: "PipelineStep",
    ):
        """Add a subprocessing step to the pipeline step."""
        self.pipeline_steps.append(pipeline_step)

    def to_dict(self) -> dict:
        """Return the processing step as a dictionary."""
        return {
            "model_slug": self.model_slug,
            "sparse_sampling_step": self.sparse_sampling_step,
            "pipeline_steps": [
                pipeline_step.to_dict() for pipeline_step in self.pipeline_steps
            ],
        }

    def run(
        self,
        sequence_input: (
            SeqRecord
            | list[SeqRecord]
            | SeqIO.FastaIO.FastaIterator
            | SeqIO.QualityIO.FastqPhredIterator
            | Path
        ),
    ) -> ModelResult:
        """Run the model on a given input."""
        model = get_model_by_slug(self.model_slug)
        model_result = model.predict(sequence_input, step=self.sparse_sampling_step)

        for pipeline_step in self.pipeline_steps:
            if pipeline_step.subprocessing_type == StepType.PREDICTION:
                score = model_result.get_scores()["total"][pipeline_step.label]
                if score >= pipeline_step.treshold:
                    prediction_model_result = pipeline_step.model_execution.run(
                        sequence_input
                    )
                    subprocessing_step = SubprocessingStep(
                        pipeline_step.subprocessing_type,
                        pipeline_step.label,
                        pipeline_step.treshold,
                        prediction_model_result,
                    )
                    model_result.add_subprocessing_step(subprocessing_step)
            elif pipeline_step.subprocessing_type == StepType.FILTERING:
                filtered_sequence_ids = model_result.get_filtered_subsequences(
                    pipeline_step.label, pipeline_step.treshold
                )
                sequence_input = get_records_by_id(
                    sequence_input, filtered_sequence_ids
                )

                filtering_model_result = None
                if sequence_input:
                    filtering_model_result = pipeline_step.model_execution.run(
                        sequence_input
                    )

                subprocessing_step = SubprocessingStep(
                    pipeline_step.subprocessing_type,
                    pipeline_step.label,
                    pipeline_step.treshold,
                    filtering_model_result,
                )
                model_result.add_subprocessing_step(subprocessing_step)
            else:
                raise ValueError(
                    f"Invalid subprocessing type {pipeline_step.subprocessing_type}"
                )

        return model_result


class PipelineStep:
    """Class for storing a subprocessing step of an XspecT model."""

    def __init__(
        self,
        subprocessing_type: StepType,
        label: str,
        treshold: float,
        model_execution: ModelExecution,
    ):
        self.subprocessing_type = subprocessing_type
        self.label = label
        self.treshold = treshold
        self.model_execution = model_execution

    def to_dict(self) -> dict:
        """Return the subprocessing step as a dictionary."""
        return {
            "subprocessing_type": str(self.subprocessing_type),
            "label": self.label,
            "treshold": self.treshold,
            "model_execution": self.model_execution.to_dict(),
        }


class Pipeline:
    """Class for storing an XspecT pipeline consisting of multiple model processing steps."""

    def __init__(self, display_name: str, author: str, author_email: str):
        self.display_name = display_name
        self.author = author
        self.author_email = author_email
        self.model_executions = []

    def add_pipeline_step(
        self,
        pipeline_step: ModelExecution,
    ):
        """Add a processing step to the pipeline."""
        self.model_executions.append(pipeline_step)

    def to_dict(self) -> dict:
        """Return the pipeline as a dictionary."""
        return {
            "display_name": self.display_name,
            "author": self.author,
            "author_email": self.author_email,
            "model_executions": [
                model_execution.to_dict() for model_execution in self.model_executions
            ],
        }

    def to_json(self) -> str:
        """Return the pipeline as a JSON string."""
        return json.dumps(self.to_dict())

    def save(self, path: Path) -> None:
        """Save the pipeline as a JSON file."""
        with open(path, "w", encoding="utf-8") as f:
            f.write(self.to_json())

    @staticmethod
    def from_file(path: Path) -> "Pipeline":
        """Load the pipeline from a JSON file."""
        with open(path, "r", encoding="utf-8") as f:
            pipeline_json = json.load(f)
            pipeline = Pipeline(
                pipeline_json["display_name"],
                pipeline_json["author"],
                pipeline_json["author_email"],
            )
            for model_execution in pipeline_json["model_executions"]:
                model_execution = ModelExecution(
                    model_execution["model_slug"],
                    model_execution["sparse_sampling_step"],
                )
                for pipeline_step in model_execution["pipeline_steps"]:
                    model_execution.add_pipeline_step(
                        PipelineStep(
                            StepType(pipeline_step["subprocessing_type"]),
                            pipeline_step["label"],
                            pipeline_step["treshold"],
                            ModelExecution(
                                pipeline_step["model_execution"]["model_slug"],
                                pipeline_step["model_execution"][
                                    "sparse_sampling_step"
                                ],
                            ),
                        )
                    )
                pipeline.add_pipeline_step(model_execution)
            return pipeline

    def run(self, input_file: Path) -> Run:
        """Run the pipeline on a given input."""
        run = Run(self.display_name, input_file)

        for model_execution in self.model_executions:
            result = model_execution.run(input_file)
            run.add_result(result)

        return run



"""This module contains functions to select and download assemblies for SVM creation."""

from time import sleep
from loguru import logger
from xspect.train_filter.ncbi_api import download_assemblies


def select_assemblies(accessions):
    """Selects up to 4 assemblies, ideally assemblies that were not used for training the filters.

    :param accessions: All selected assembly accessions for every species.
    :type accessions: dict
    :return: Dict with species name as key and selected accessions as value.
    """

    all_accessions = {
        sci_name: curr_accessions[-4:]
        for sci_name, curr_accessions in accessions.items()
    }

    return all_accessions


def get_svm_assemblies(all_accessions, dir_name):
    """Download assemblies for svm creation.

    :param all_accessions: Contains lists with all previously selected assemblies for every species.
    :type all_accessions: dict
    :param dir_name: Name of the parent directory.
    :type dir_name: str
    """
    # Select accessions for download.
    selected_accessions = select_assemblies(all_accessions)

    # Download assemblies.
    for sci_name, accessions in selected_accessions.items():
        sleep(5)
        logger.info("Downloading {name}", name=sci_name)
        file_name = sci_name + ".zip"
        download_assemblies.download_assemblies(
            accessions=accessions,
            dir_name=dir_name,
            target_folder="training_data_zipped",
            zip_file_name=file_name,
        )



"""This class uses the NCBI Datasets API to get the taxonomy tree of a given Taxon.

The taxonomy tree consists of only the next children to the parent taxon.
The children are only of the next lower rank of the parent taxon.
"""

__author__ = "Berger, Phillip"

import sys
import requests

from loguru import logger


class NCBIChildrenTree:
    _taxon: str
    _response: dict
    _parent_taxon_id: str
    _children_taxon_ids = list()

    def __init__(self, taxon: str):
        self._taxon = taxon
        self._request_tree()

    def _request_tree(self):
        """Make the request for the children tree at the NCBI Datasets API."""
        api_url = f"https://api.ncbi.nlm.nih.gov/datasets/v1/taxonomy/taxon/{self._taxon}/filtered_subtree"
        raw_response = requests.get(api_url, timeout=5)
        self._response = raw_response.json()["edges"]
        self._parent_taxon_id = str(self._response["1"]["visible_children"][0])
        try:
            tmp_children_ids = self._response[self._parent_taxon_id]["visible_children"]
        except KeyError:
            logger.error("KeyError for key: {key}", key=self._parent_taxon_id)
            logger.error("Response: {response}", response=str(self._response))
            logger.error("Aborting")
            sys.exit()
        for child_id in tmp_children_ids:
            self._children_taxon_ids.append(str(child_id))

    def parent_id(self):
        """The NCBI taxon ID of the given Taxon at the initialisation.

        :return: The taxon ID.
        """
        return self._parent_taxon_id

    def children_ids(self) -> list[str]:
        """The NCBI taxon IDs of all children of the given Taxon. The children are all of a lower rank than the parent.

        :return: The taxon IDs as a list.
        """
        return self._children_taxon_ids



"""
File IO module tests.
"""

import os
from pathlib import Path
from src.xspect.file_io import (
    concatenate_meta,
)


def test_concatenate_meta(tmpdir, monkeypatch):
    """Test if the function concatenates fasta files correctly."""
    # Set up temporary directory
    monkeypatch.chdir(tmpdir)

    # Create a temporary directory for the concatenated fasta files
    concatenate_dir = Path(tmpdir) / "concatenate"
    concatenate_dir.mkdir()

    # Create some temporary fasta files
    fasta_files = [
        "file1.fasta",
        "file2.fna",
        "file3.fa",
        "file4.ffn",
        "file5.frn",
        "file6.txt",
        "file7.jpg",
        "file8.png",
    ]
    for file in fasta_files:
        with open(concatenate_dir / file, "w", encoding="utf-8") as f:
            f.write(f">{file}\n{file}")

    # Call the function to be tested
    concatenate_meta(tmpdir, "Salmonella")

    # Check if the meta file has been created and contains the correct content
    meta_file = Path(tmpdir) / "Salmonella.fasta"
    assert meta_file.is_file()

    with open(meta_file, "r", encoding="utf-8") as f:
        content = f.read()
        assert content.startswith(">Salmonella metagenome")
        for file in fasta_files:
            if (
                file.endswith(".fasta")
                or file.endswith(".fna")
                or file.endswith(".fa")
                or file.endswith(".ffn")
                or file.endswith(".frn")
            ):
                assert file in content
            else:
                assert file not in content



""" Module with XspecT global run class, which summarizes individual model results. """

import json
from pathlib import Path
from xspect.models.result import ModelResult


class Run:
    """Class for storing the results of an XspecT run."""

    def __init__(self, display_name: str, input_file: str):
        self.display_name = display_name
        self.input_file = input_file
        self.results = []

    def add_result(self, result: ModelResult):
        """Add a result to the run."""
        self.results.append(result)

    def to_dict(self) -> dict:
        """Return the run as a dictionary."""
        return {
            "display_name": self.display_name,
            "input_file": str(self.input_file),
            "results": (
                [result.to_dict() for result in self.results] if self.results else []
            ),
        }

    def to_json(self) -> str:
        """Return the run as a JSON string."""
        json_dict = self.to_dict()
        return json.dumps(json_dict, indent=4)

    def save(self, path: Path) -> None:
        """Save the run as a JSON file."""
        with open(path, "w", encoding="utf-8") as f:
            f.write(self.to_json())



"""Probabilistic filter SVM model for sequence data"""

# pylint: disable=no-name-in-module, too-many-instance-attributes

import json
from math import ceil
from pathlib import Path
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from rbloom import Bloom
from xxhash import xxh3_64_intdigest
from xspect.models.probabilistic_filter_model import ProbabilisticFilterModel
from xspect.file_io import get_record_iterator


class ProbabilisticSingleFilterModel(ProbabilisticFilterModel):
    """Probabilistic filter SVM model for sequence data"""

    def __init__(
        self,
        k: int,
        model_display_name: str,
        author: str,
        author_email: str,
        model_type: str,
        base_path: Path,
        fpr: float = 0.01,
        num_hashes: int = 7,
    ) -> None:
        super().__init__(
            k=k,
            model_display_name=model_display_name,
            author=author,
            author_email=author_email,
            model_type=model_type,
            base_path=base_path,
            fpr=fpr,
            num_hashes=num_hashes,
        )
        self.bf = None

    def fit(self, file_path: Path, display_name: str) -> None:
        """Fit the SVM to the sequences and labels"""
        # estimate number of kmers
        total_length = 0
        for record in get_record_iterator(file_path):
            total_length += len(record.seq)
        num_kmers = total_length - self.k + 1

        self.bf = Bloom(num_kmers, self.fpr, hash_func=xxh3_64_intdigest)
        for record in get_record_iterator(file_path):
            for kmer in self._generate_kmers(record.seq):
                self.bf.add(kmer)
        self.display_names[file_path.stem] = display_name

        bloom_path = self.base_path / self.slug() / "filter.bloom"
        bloom_path.parent.mkdir(parents=True, exist_ok=True)
        self.bf.save(str(bloom_path))

    def calculate_hits(
        self, sequence: Seq | SeqRecord, filter_ids=None, step: int = 1
    ) -> dict:
        """Calculate the hits for the sequence"""
        if isinstance(sequence, SeqRecord):
            sequence = sequence.seq

        if not isinstance(sequence, Seq):
            raise ValueError("Invalid sequence, must be a Bio.Seq object")

        if not len(sequence) > self.k:
            raise ValueError("Invalid sequence, must be longer than k")

        num_hits = sum(
            1 for kmer in self._generate_kmers(sequence, step=step) if kmer in self.bf
        )
        return {next(iter(self.display_names)): num_hits}

    @staticmethod
    def load(path: Path) -> "ProbabilisticSingleFilterModel":
        """Load the model from disk"""
        with open(path, "r", encoding="utf-8") as file:
            json_object = file.read()
            model_json = json.loads(json_object)
            model = ProbabilisticSingleFilterModel(
                model_json["k"],
                model_json["model_display_name"],
                model_json["author"],
                model_json["author_email"],
                model_json["model_type"],
                path.parent,
                fpr=model_json["fpr"],
                num_hashes=model_json["num_hashes"],
            )
            model.display_names = model_json["display_names"]
            bloom_path = model.base_path / model.slug() / "filter.bloom"
            model.bf = Bloom.load(
                str(bloom_path),
                hash_func=xxh3_64_intdigest,
            )
            return model

    def _generate_kmers(self, sequence: Seq, step: int = 1):
        """Generate kmers from the sequence"""
        num_kmers = ceil((len(sequence) - self.k + 1) / step)
        for i in range(num_kmers):
            start_pos = i * step
            kmer = sequence[start_pos : start_pos + self.k]
            minimizer = min(kmer, str(kmer.reverse_complement()))
            yield str(minimizer)



""" Module for storing the results of XspecT models. """

from enum import Enum


def get_last_processing_step(result: "ModelResult") -> "ModelResult":
    """Get the last subprocessing step of the result. First path only."""
    last_step = result
    while last_step.subprocessing_steps:
        last_step = last_step.subprocessing_steps[-1].result
    return last_step


class StepType(Enum):
    """Enum for defining the type of a subprocessing step."""

    PREDICTION = 1
    FILTERING = 2

    def __str__(self) -> str:
        return self.name.lower()


class SubprocessingStep:
    """Class for storing a subprocessing step of an XspecT model."""

    def __init__(
        self,
        subprocessing_type: StepType,
        label: str,
        treshold: float,
        result: "ModelResult",
    ):
        self.subprocessing_type = subprocessing_type
        self.label = label
        self.treshold = treshold
        self.result = result

    def to_dict(self) -> dict:
        """Return the subprocessing step as a dictionary."""
        return {
            "subprocessing_type": str(self.subprocessing_type),
            "label": self.label,
            "treshold": self.treshold,
            "result": self.result.to_dict() if self.result else {},
        }


class ModelResult:
    """Class for storing an XspecT model result."""

    def __init__(
        self,
        # we store hits depending on the subsequence as well as on the label
        model_slug: str,
        hits: dict[str, dict[str, int]],
        num_kmers: dict[str, int],
        sparse_sampling_step: int = 1,
        prediction: str = None,
    ):
        if "total" in hits:
            raise ValueError(
                "'total' is a reserved key and cannot be used as a subsequence"
            )
        self.model_slug = model_slug
        self.hits = hits
        self.num_kmers = num_kmers
        self.sparse_sampling_step = sparse_sampling_step
        self.prediction = prediction
        self.subprocessing_steps = []

    def add_subprocessing_step(self, subprocessing_step: SubprocessingStep) -> None:
        """Add a subprocessing step to the result."""
        if subprocessing_step.label in self.subprocessing_steps:
            raise ValueError(
                f"Subprocessing step {subprocessing_step.label} already exists in the result"
            )
        self.subprocessing_steps.append(subprocessing_step)

    def get_scores(self) -> dict:
        """Return the scores of the model."""
        scores = {
            subsequence: {
                label: round(hits / self.num_kmers[subsequence], 2)
                for label, hits in subseuqence_hits.items()
            }
            for subsequence, subseuqence_hits in self.hits.items()
        }

        # calculate total scores
        total_num_kmers = sum(self.num_kmers.values())
        total_hits = self.get_total_hits()

        scores["total"] = {
            label: round(hits / total_num_kmers, 2)
            for label, hits in total_hits.items()
        }

        return scores

    def get_total_hits(self) -> dict[str, int]:
        """Return the total hits of the model."""
        total_hits = {label: 0 for label in list(self.hits.values())[0]}
        for _, subseuqence_hits in self.hits.items():
            for label, hits in subseuqence_hits.items():
                total_hits[label] += hits
        return total_hits

    def get_filter_mask(self, label: str, filter_threshold: float) -> dict[str, bool]:
        """Return a mask for filtered subsequences."""
        if filter_threshold < 0 or filter_threshold > 1:
            raise ValueError("The filter threshold must be between 0 and 1.")

        scores = self.get_scores()
        scores.pop("total")
        return {
            subsequence: score[label] >= filter_threshold
            for subsequence, score in scores.items()
        }

    def get_filtered_subsequences(self, label: str, filter_threshold: 0.7) -> list[str]:
        """Return the filtered subsequences."""
        return [
            subsequence
            for subsequence, mask in self.get_filter_mask(
                label, filter_threshold
            ).items()
            if mask
        ]

    def to_dict(self) -> dict:
        """Return the result as a dictionary."""
        res = {
            "model_slug": self.model_slug,
            "sparse_sampling_step": self.sparse_sampling_step,
            "hits": self.hits,
            "scores": self.get_scores(),
            "num_kmers": self.num_kmers,
            "subprocessing_steps": [
                subprocessing_step.to_dict()
                for subprocessing_step in self.subprocessing_steps
            ],
        }

        if self.prediction is not None:
            res["prediction"] = self.prediction

        return res



"""
This module contains the main functions for training the models.
"""

import os
import shutil
from pathlib import Path
import sys
from time import localtime, perf_counter, asctime, sleep
from loguru import logger
from xspect.definitions import get_xspect_model_path, get_xspect_tmp_path
from xspect.file_io import concatenate_meta
from xspect.models.probabilistic_filter_svm_model import ProbabilisticFilterSVMModel
from xspect.models.probabilistic_single_filter_model import (
    ProbabilisticSingleFilterModel,
)
from xspect.train_filter.ncbi_api import (
    ncbi_assembly_metadata,
    ncbi_taxon_metadata,
    ncbi_children_tree,
    download_assemblies,
)
from xspect.train_filter import (
    create_svm,
    html_scrap,
    extract_and_concatenate,
)


def check_user_input(user_input: str):
    """The given input of the user will be checked. The input has to be a genus in NCBI.

    :return: The genus name.
    """
    taxon_metadata = ncbi_taxon_metadata.NCBITaxonMetadata([user_input])
    all_metadata = taxon_metadata.get_metadata()
    for metadata in all_metadata.values():
        sci_name = metadata["sci_name"]
        tax_id = metadata["tax_id"]
        rank = metadata["rank"]
        lineage = metadata["lineage"]
        bacteria_id = 2
        if not sci_name == user_input and not tax_id == user_input:
            print(
                f"{get_current_time()}| The given genus: {user_input} was found as"
                f" genus: {sci_name} ID: {tax_id}"
            )
            print(f"{get_current_time()}| Using {sci_name} as genus name.")
        if rank == "GENUS":
            if bacteria_id not in lineage:
                print(f"{get_current_time()}| The given genus is not a bacteria.")
                print(f"{get_current_time()}| Do you want to continue: [y/n]")
                choice = input("-> ").lower()
                if choice == "y":
                    return str(sci_name)
                print(f"{get_current_time()}| Exiting...")
                sys.exit()
            return str(sci_name)
        print(f"{get_current_time()}| {user_input} is rank {rank} and not genus.")
        sys.exit()


def copy_custom_data(bf_path: str, svm_path: str, dir_name: str):
    """

    :param bf_path:
    :param svm_path:
    :param dir_name:
    :return:
    """
    path = Path(os.getcwd()) / "genus_metadata" / dir_name
    new_bf_path = path / "concatenate"
    new_svm_path = path / "training_data"

    # Make the new directories.
    path.mkdir(exist_ok=True)
    new_bf_path.mkdir(exist_ok=True)
    new_svm_path.mkdir(exist_ok=True)

    # Move bloomfilter files.
    bf_files = os.listdir(bf_path)
    for file in bf_files:
        file_path = Path(bf_path) / file
        new_file_path = new_bf_path / file
        shutil.copy2(file_path, new_file_path)

    # Move svm files.
    svm_files = os.listdir(svm_path)
    for file in svm_files:
        file_path = Path(svm_path) / file
        new_file_path = new_svm_path / file
        shutil.copy2(file_path, new_file_path)


def set_logger(dir_name: str):
    """Sets the logger parameters.

    :param dir_name: Name of the folder where the log should be saved.
    """
    genus = dir_name.split("_")[0]

    # Starting logger.
    logger.remove()
    logger.add(sys.stderr, format="{time:HH:mm:ss} | {level} | {message}", level="INFO")
    log_path = get_xspect_tmp_path() / dir_name / (genus + ".log")
    logger.add(log_path, format="{time:HH:mm:ss} | {level} | {message}", level="DEBUG")


def create_translation_dict(dir_name: str) -> dict[str, str]:
    """Create a translation dictionary to translate the taxon ID to its scientific name.

    :param dir_name: Directory name for current genus.
    :return: The created translation dictionary.
    """
    path = get_xspect_tmp_path() / dir_name / "concatenate"
    files = os.listdir(path)
    translation_dict = {}
    for file in files:
        file_split = file.split(".")[0].split("_")
        tax_id = file_split[0]
        final_file_name = tax_id + ".fasta"
        name = file_split[1]
        translation_dict[final_file_name] = name

    return translation_dict


def change_bf_assembly_file_names(dir_name: str):
    """Change all concatenated assembly names to only the taxon ID.

    :param dir_name: Directory name for current genus.
    """
    path = get_xspect_tmp_path() / dir_name / "concatenate"
    files = os.listdir(path)
    for file in files:
        file_split = file.split(".")[0].split("_")
        tax_id = file_split[0]
        new_file_name = f"{tax_id}.fasta"
        os.rename((path / file), (path / new_file_name))


def get_current_time():
    """Returns the current time in the form hh:mm:ss."""
    return asctime(localtime()).split()[3]


def train_ncbi(genus: str, svm_step: int = 1):
    """Train genus and species models with NCBI assemblies from the given genus."""

    if not isinstance(genus, str):
        raise TypeError("genus must be a string")

    # Check user input.
    genus = check_user_input(user_input=genus)

    # The directory name is defined in the following format: 'genus'_DD_MM_YYYY_hh-mm-ss
    curr_time = localtime()
    dir_name = f"{genus}_{curr_time[2]}_{curr_time[1]}_{curr_time[0]}_{curr_time[3]}-{curr_time[4]}-{curr_time[5]}"

    # Set the logger.
    set_logger(dir_name)

    # Time for the whole program.
    start = perf_counter()

    # Search for every defined species of the genus.
    logger.info("Getting all species of the genus")
    children_ids = ncbi_children_tree.NCBIChildrenTree(genus).children_ids()
    species_dict = ncbi_taxon_metadata.NCBITaxonMetadata(children_ids).get_metadata()

    # Get all gcf accessions that have Taxonomy check result OK.
    logger.info("Checking ANI data for updates")
    ani_gcf = html_scrap.TaxonomyCheck().ani_gcf()

    # Look for up to 8 assembly accessions per species.
    logger.info("Getting assembly metadata")
    all_metadata = ncbi_assembly_metadata.NCBIAssemblyMetadata(
        all_metadata=species_dict, ani_gcf=ani_gcf, count=8, contig_n50=10000
    )
    all_metadata = all_metadata.get_all_metadata()

    # Ensure that the genus has at least one species with accessions.
    if not all_metadata:
        raise ValueError("No species with accessions found")

    # Download the chosen assemblies.
    # One file for each species with it's downloaded assemblies in zip format.

    # Iterate through all species.
    logger.info("Downloading assemblies for bloomfilter training")
    for metadata in all_metadata.values():
        # Only try to download when the species has accessions.
        if len(metadata["accessions"]) >= 1:
            sleep(5)
            species_name = metadata["sci_name"]
            tax_id = metadata["tax_id"]
            logger.info("Downloading {id}_{name}", id=tax_id, name=species_name)
            file_name = f"{tax_id}_{species_name}.zip"

            # Selecting the first 4 assemblies for training the filters.
            accessions = metadata["accessions"][:4]

            download_assemblies.download_assemblies(
                accessions=accessions,
                dir_name=dir_name,
                target_folder="zip_files",
                zip_file_name=file_name,
            )
    logger.info("Concatenating and extracting")

    # Concatenate all assemblies of each species.
    extract_and_concatenate.bf(dir_name=dir_name, delete=True)
    concatenate_meta(get_xspect_tmp_path() / dir_name, genus)

    # Download assemblies for svm creation.
    logger.info("Downloading assemblies for support-vector-machine training")
    accessions = {}
    for metadata in all_metadata.values():
        # Only add taxon with accessions.
        if len(metadata["accessions"]) >= 1:
            accessions[metadata["tax_id"]] = metadata["accessions"]

    # Downloading assemblies.
    create_svm.get_svm_assemblies(all_accessions=accessions, dir_name=dir_name)

    logger.info("Extracting SVM assemblies")

    # Extracting assemblies.
    extract_and_concatenate.svm(
        species_accessions=accessions, dir_name=dir_name, delete=True
    )

    # Make dictionary for translating taxon ID to scientific name.
    translation_dict = create_translation_dict(dir_name)
    change_bf_assembly_file_names(dir_name)

    species_files_path = get_xspect_tmp_path() / dir_name / "concatenate"
    species_result_path = get_xspect_model_path() / genus

    # Train Bloomfilter for complete genus.
    logger.info("Training metagenome model")
    mg_files_path = get_xspect_tmp_path() / dir_name

    genus_model = ProbabilisticSingleFilterModel(
        k=21,
        model_display_name=genus,
        author="Test",
        author_email="test@example.com",
        model_type="Genus",
        base_path=Path(species_result_path).parent,
    )
    genus_model.fit(mg_files_path / f"{genus}.fasta", genus)
    genus_model.save()

    logger.info("Training species model")

    species_model = ProbabilisticFilterSVMModel(
        k=21,
        model_display_name=genus,
        author="Test",
        author_email="test@example.com",
        model_type="Species",
        base_path=Path(species_result_path).parent,
        kernel="rbf",
        c=1.0,
    )
    svm_dir = get_xspect_tmp_path() / dir_name / "training_data"
    species_model.fit(
        Path(species_files_path),
        svm_dir,
        display_names=translation_dict,
        svm_step=svm_step,
    )
    species_model.save()

    # Cleanup files.
    shutil.rmtree(get_xspect_tmp_path() / dir_name)

    end = perf_counter()

    logger.info("Program runtime: {time} m", time=(round((end - start) / 60, 2)))
    logger.info("XspecT-trainer is finished.")


def train_from_directory(display_name: str, dir_path: Path, meta: bool = False):
    """Train the gene family and gene filter.

    :param display_name: Name of the model.
    :param dir: Input directory.
    """

    if not isinstance(display_name, str):
        raise TypeError("display_name must be a string")

    if not isinstance(dir_path, Path) and dir_path.exists() and dir_path.is_dir():
        raise ValueError("dir must be Path object to a valid directory")

    # check if the directory contains the necessary files
    # copy to temp path
    # check if svm training data exists
    # train model, with svm data if it exists
    # add display names
    # train metagenome model
    # clean up temp path



"""
Training script tests
"""

import pytest
from src.xspect import train


def test_invalid_taxonomy_check():
    """
    Test if a ValueError is thrown when attempting to train a genus
    where species do not fulfill taxonomy check requirements.
    """
    with pytest.raises(ValueError):
        train.train_ncbi("Amnimonas")



""" Collects metadata of assemblies from NCBI API """

__author__ = "Berger, Phillip"

from time import sleep

import requests

from loguru import logger


class NCBIAssemblyMetadata:
    """Class to collect metadata of assemblies from the NCBI API."""

    _all_metadata: dict
    _count: int
    _ani_gcf: list
    _parameters: dict
    _accessions: list[str]
    _contig_n50: int
    _all_metadata_complete: dict

    def __init__(self, all_metadata: dict, ani_gcf: list, count=8, contig_n50=10000):
        self._all_metadata = all_metadata
        self._count = count
        self._ani_gcf = ani_gcf
        self._contig_n50 = contig_n50

        self._set_parameters()

        tmp_metadata = {}
        for tax_id, curr_metadata in self._all_metadata.items():
            sleep(2)
            species_name = curr_metadata["sci_name"]
            logger.info("Collecting metadata of {name}", name=species_name)
            accessions = self._make_request(taxon=tax_id)
            if len(accessions) != 0:
                curr_metadata["accessions"] = accessions
                tmp_metadata[tax_id] = curr_metadata

        self._all_metadata_complete = tmp_metadata

    def _set_parameters(self):
        params = {
            "filters.reference_only": "false",
            "filters.assembly_source": "refseq",
            "filters.exclude_atypical": "true",
            "page_size": self._count,
            "page_token": "",
        }
        params_ref = params.copy()
        params_ref["filters.reference_only"] = "true"

        params_comp_genome = params.copy()
        params_comp_genome["filters.assembly_level"] = "complete_genome"

        params_chrom = params.copy()
        params_chrom["filters.assembly_level"] = "chromosome"

        params_scaffold = params.copy()
        params_scaffold["filters.assembly_level"] = "scaffold"

        params_contig = params.copy()
        params_contig["filters.assembly_level"] = "contig"

        self._parameters = {
            "params_ref": params_ref,
            "params_comp_genome": params_comp_genome,
            "params_chrom": params_chrom,
            "params_scaffold": params_scaffold,
            "params_contig": params_contig,
        }

    def _make_request(self, taxon: str):
        api_url = f"https://api.ncbi.nlm.nih.gov/datasets/v1/genome/taxon/{taxon}"
        accessions = []
        count = 0
        for request_type, parameters in self._parameters.items():
            raw_response = requests.get(api_url, params=parameters, timeout=5)
            response = raw_response.json()
            if response:
                try:
                    assemblies = response["assemblies"]
                    for assembly in assemblies:
                        curr_assembly = assembly["assembly"]
                        curr_accession = curr_assembly["assembly_accession"]
                        curr_contig_n50 = curr_assembly["contig_n50"]
                        if count < self._count:
                            if (
                                curr_accession in self._ani_gcf
                                and curr_contig_n50 > self._contig_n50
                            ):
                                accessions.append(curr_accession)
                                count += 1
                        else:
                            break
                except KeyError:
                    logger.debug(
                        "While requesting: {type} an error response was given",
                        type=request_type,
                    )
                    logger.debug(str(response))

            if count >= self._count:
                break
        return accessions

    def get_all_metadata(self):
        """Returns all metadata of the assemblies."""
        return self._all_metadata_complete



"""FastAPI application for XspecT."""

import datetime
from pathlib import Path
from shutil import copyfileobj
from fastapi import FastAPI, UploadFile, BackgroundTasks
from xspect.definitions import get_xspect_runs_path, get_xspect_upload_path
from xspect.download_filters import download_test_filters
import xspect.model_management as mm
from xspect.models.result import StepType
from xspect.pipeline import ModelExecution, Pipeline, PipelineStep
from xspect.train import train_ncbi

app = FastAPI()


@app.get("/download-filters")
def download_filters():
    """Download filters."""
    download_test_filters("https://xspect2.s3.eu-central-1.amazonaws.com/models.zip")


@app.get("/classify")
def classify(genus: str, file: str, meta: bool = False, step: int = 500):
    """Classify uploaded sample."""

    path = get_xspect_upload_path() / file

    pipeline = Pipeline(genus + " classification", "Test Author", "test@example.com")
    species_execution = ModelExecution(genus + "-species", sparse_sampling_step=step)
    if meta:
        species_filtering_step = PipelineStep(
            StepType.FILTERING, genus, 0.7, species_execution
        )
        genus_execution = ModelExecution(genus + "-genus", sparse_sampling_step=step)
        genus_execution.add_pipeline_step(species_filtering_step)
        pipeline.add_pipeline_step(genus_execution)
    else:
        pipeline.add_pipeline_step(species_execution)

    run = pipeline.run(Path(path))
    time_str = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
    save_path = get_xspect_runs_path() / f"run_{time_str}.json"
    run.save(save_path)

    return run.to_dict()


@app.post("/train")
def train(genus: str, background_tasks: BackgroundTasks, svm_steps: int = 1):
    """Train NCBI model."""
    background_tasks.add_task(train_ncbi, genus, svm_steps)

    return {"message": "Training started."}


@app.get("/list-models")
def list_models():
    """List available models."""
    return mm.get_models()


@app.get("/model-metadata")
def get_model_metadata(model_slug: str):
    """Get metadata of a model."""
    return mm.get_model_metadata(model_slug)


@app.post("/model-metadata")
def post_model_metadata(model_slug: str, author: str, author_email: str):
    """Update metadata of a model."""
    try:
        mm.update_model_metadata(model_slug, author, author_email)
    except ValueError as e:
        return {"error": str(e)}
    return {"message": "Metadata updated."}


@app.post("/model-display-name")
def post_model_display_name(model_slug: str, filter_id: str, display_name: str):
    """Update display name of a filter in a model."""
    try:
        mm.update_model_display_name(model_slug, filter_id, display_name)
    except ValueError as e:
        return {"error": str(e)}
    return {"message": "Display name updated."}


@app.post("/upload-file")
def upload_file(file: UploadFile):
    """Upload file to the server."""
    upload_path = get_xspect_upload_path() / file.filename

    if not upload_path.exists():
        try:
            with upload_path.open("wb") as buffer:
                copyfileobj(file.file, buffer)
        finally:
            file.file.close()

    return {"filename": file.filename}



"""Tests for the ProbabilisticFilterModel class."""

# pylint: disable=redefined-outer-name, line-too-long, protected-access

from pathlib import Path
import pytest
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from src.xspect.models.probabilistic_filter_model import ProbabilisticFilterModel


@pytest.fixture
def filter_model(tmpdir):
    """Fixture for the ProbabilisticFilterModel class."""
    base_path = tmpdir.mkdir("xspect_data")
    return ProbabilisticFilterModel(
        k=21,
        model_display_name="Test Filter",
        author="John Doe",
        author_email="john.doe@example.com",
        model_type="Species",
        base_path=Path(base_path),
    )


@pytest.fixture
def trained_filter_model(filter_model, multiple_assembly_dir_path):
    """Fixture for the ProbabilisticFilterModel class with trained model."""
    filter_model.fit(Path(multiple_assembly_dir_path))
    return filter_model


def test_model_initialization(filter_model):
    """Test the initialization of the ProbabilisticFilterModel class."""
    assert filter_model.k == 21
    assert filter_model.model_display_name == "Test Filter"
    assert filter_model.author == "John Doe"
    assert filter_model.author_email == "john.doe@example.com"
    assert filter_model.model_type == "Species"


def test_fit(filter_model, multiple_assembly_dir_path):
    """Test the fit method of the ProbabilisticFilterModel class."""
    filter_model.fit(Path(multiple_assembly_dir_path))
    cobs_path = filter_model.base_path / filter_model.slug() / "index.cobs_classic"
    assert cobs_path.exists()
    assert filter_model.display_names == {
        "GCF_000006945": "GCF_000006945.2_ASM694v2_genomic",
        "GCF_000018445": "GCF_000018445.1_ASM1844v1_genomic",
        "GCF_000069245": "GCF_000069245.1_ASM6924v1_genomic",
    }


def test_fit_with_display_names(filter_model, multiple_assembly_dir_path):
    """Test the fit method of the ProbabilisticFilterModel class with display names."""
    display_names = {
        "GCF_000006945.2_ASM694v2_genomic.fna": "Salmonella enterica",
        "GCF_000018445.1_ASM1844v1_genomic.fna": "Acinetobacter baumannii ACICU",
        "GCF_000069245.1_ASM6924v1_genomic.fna": "Acinetobacter baumannii AYE",
    }
    # cobs only uses the names until the first "." internally
    expected_display_names = {
        "GCF_000006945": "Salmonella enterica",
        "GCF_000018445": "Acinetobacter baumannii ACICU",
        "GCF_000069245": "Acinetobacter baumannii AYE",
    }
    filter_model.fit(Path(multiple_assembly_dir_path), display_names=display_names)
    cobs_path = filter_model.base_path / filter_model.slug() / "index.cobs_classic"
    assert cobs_path.exists()
    assert filter_model.display_names == expected_display_names


def test_predict(trained_filter_model):
    """Test the predict method of the ProbabilisticFilterModel class."""

    salmonella_record = SeqRecord(
        seq=Seq(
            "AGAGATTACGTCTGGTTGCAAGAGATCATGACAGGGGGAATTGGTTGAAAATAAATATATCGCCAGCAGCACATGAACAA"
        ),
        id="test",
    )

    res = trained_filter_model.predict(salmonella_record)
    assert res.get_scores()["total"] == {
        "GCF_000006945": 1.0,
        "GCF_000069245": 0.0,
        "GCF_000018445": 0.0,
    }
    assert res.get_total_hits() == {
        "GCF_000006945": 60,
        "GCF_000069245": 0,
        "GCF_000018445": 0,
    }


@pytest.mark.parametrize(
    ["assembly_file_path", "expected_scores"],
    [
        (
            "GCF_000069245.1_ASM6924v1_genomic.fna",
            {"GCF_000006945": 0.01, "GCF_000018445": 0.63, "GCF_000069245": 1.0},
        ),
        (
            "GCF_000018445.1_ASM1844v1_genomic.fna",
            {"GCF_000006945": 0.01, "GCF_000018445": 1.0, "GCF_000069245": 0.64},
        ),
        (
            "GCF_000006945.2_ASM694v2_genomic.fna",
            {"GCF_000006945": 1.0, "GCF_000018445": 0.0, "GCF_000069245": 0.0},
        ),
    ],
    indirect=["assembly_file_path"],
)
def test_predict_from_file(trained_filter_model, assembly_file_path, expected_scores):
    """Test the predict method of the ProbabilisticFilterModel class with a file
    containing multiple (sub)sequences."""
    res = trained_filter_model.predict(Path(assembly_file_path))
    assert res.get_scores()["total"] == expected_scores


def test_model_save_and_load(trained_filter_model):
    """Test the save and load methods of the ProbabilisticFilterModel class."""
    trained_filter_model.save()
    loaded_model = ProbabilisticFilterModel.load(
        trained_filter_model.base_path / "test-filter-species.json"
    )
    assert trained_filter_model.to_dict() == loaded_model.to_dict()


def test_trained_filter_model_save_and_load(trained_filter_model):
    """Test the save and load methods of the ProbabilisticFilterModel class."""
    trained_filter_model.save()
    loaded_model = ProbabilisticFilterModel.load(
        trained_filter_model.base_path / "test-filter-species.json"
    )
    assert trained_filter_model.to_dict() == loaded_model.to_dict()


def test_count_kmers(filter_model):
    """Test the _count_kmers method of the ProbabilisticFilterModel class."""
    sequence = Seq(
        "AGAGATTACGTCTGGTTGCAAGAGATCATGACAGGGGGAATTGGTTGAAAATAAATATATCGCCAGCAGCACATGAACAA"
    )
    kmer_counts = filter_model._count_kmers(sequence)
    # k = 21
    assert kmer_counts == 60


def test_calculate_hits(trained_filter_model):
    """Test the calculate_hits method of the ProbabilisticFilterModel class."""
    salmonella_sequence = Seq(
        "AGAGATTACGTCTGGTTGCAAGAGATCATGACAGGGGGAATTGGTTGAAAATAAATATATCGCCAGCAGCACATGAACAA"
    )

    for step in range(1, 5):
        hits = trained_filter_model.calculate_hits(salmonella_sequence, step=step)
        assert hits == {
            "GCF_000006945": 60 / step,
            "GCF_000069245": 0,
            "GCF_000018445": 0,
        }



"""
File IO module.
"""

import os
from pathlib import Path
import zipfile
from Bio import SeqIO
from xspect.definitions import fasta_endings, fastq_endings


def delete_zip_files(dir_path):
    """Delete all zip files in the given directory."""
    files = os.listdir(dir_path)
    for file in files:
        if zipfile.is_zipfile(file):
            file_path = dir_path / str(file)
            os.remove(file_path)


def extract_zip(zip_path, unzipped_path):
    """Extracts all files from a directory with zip files."""
    # Make new directory.
    os.mkdir(unzipped_path)

    file_names = os.listdir(zip_path)
    for file in file_names:
        file_path = zip_path / file
        if zipfile.is_zipfile(file_path):
            with zipfile.ZipFile(file_path) as item:
                directory = unzipped_path / file.replace(".zip", "")
                item.extractall(directory)


def concatenate_meta(path: Path, genus: str):
    """Concatenates all species files to one fasta file.

    :param path: Path to the directory with the concatenated fasta files.
    :type path: Path
    :param genus: Genus name.
    :type genus: str
    """
    files_path = path / "concatenate"
    meta_path = path / (genus + ".fasta")
    files = os.listdir(files_path)

    with open(meta_path, "w", encoding="utf-8") as meta_file:
        # Write the header.
        meta_header = f">{genus} metagenome\n"
        meta_file.write(meta_header)

        # Open each concatenated species file and write the sequence in the meta file.
        for file in files:
            file_ending = str(file).rsplit(".", maxsplit=1)[-1]
            if file_ending in fasta_endings:
                with open(
                    (files_path / str(file)), "r", encoding="utf-8"
                ) as species_file:
                    for line in species_file:
                        if line[0] != ">":
                            meta_file.write(line.replace("\n", ""))


def get_record_iterator(file_path: Path):
    """Returns a record iterator for a fasta or fastq file."""
    if not isinstance(file_path, Path):
        raise ValueError("Path must be a Path object")

    if not file_path.exists():
        raise ValueError("File does not exist")

    if not file_path.is_file():
        raise ValueError("Path must be a file")

    if file_path.suffix[1:] in fasta_endings:
        return SeqIO.parse(file_path, "fasta")

    if file_path.suffix[1:] in fastq_endings:
        return SeqIO.parse(file_path, "fastq")

    raise ValueError("Invalid file format, must be a fasta or fastq file")


def get_records_by_id(file: Path, ids: list[str]):
    """Return records with the specified ids."""
    records = get_record_iterator(file)
    return [record for record in records if record.id in ids]



"""Tests for the model_management module."""

import xspect.model_management as mm


def test_get_genus_model():
    """Test the get_genus_model function."""
    aci = mm.get_genus_model("Acinetobacter")
    sal = mm.get_genus_model("Salmonella")

    assert aci.model_display_name == "Acinetobacter"
    assert aci.model_type == "Genus"

    assert sal.model_display_name == "Salmonella"
    assert sal.model_type == "Genus"


def test_get_species_model():
    """Test the get_species_model function."""
    aci = mm.get_species_model("Acinetobacter")
    sal = mm.get_species_model("Salmonella")

    assert aci.model_display_name == "Acinetobacter"
    assert aci.model_type == "Species"

    assert sal.model_display_name == "Salmonella"
    assert sal.model_type == "Species"


def test_get_model_metadata():
    """Test the get_model_metadata function."""
    aci_meta = mm.get_model_metadata("acinetobacter-genus")

    assert aci_meta["model_display_name"] == "Acinetobacter"


def test_get_models():
    """Test the get_models function."""
    model_dict = mm.get_models()
    assert "Genus" in model_dict
    assert "Species" in model_dict
    assert "Acinetobacter" in model_dict["Genus"]
    assert "Salmonella" in model_dict["Genus"]
    assert "Acinetobacter" in model_dict["Species"]
    assert "Salmonella" in model_dict["Species"]



""" This module is used to retrieve metadata from the NCBI taxonomy database. """

__author__ = "Berger, Phillip"

import requests

from loguru import logger


class NCBITaxonMetadata:
    """Class to retrieve metadata from the NCBI taxonomy database."""

    _taxon: str
    _response: dict
    _all_metadata: dict

    def __init__(self, taxon: list[str]):
        self._taxon = ",".join(taxon)
        self._all_metadata = {}
        self._request_metadata()
        self._collect_all_metadata()

    def _request_metadata(self):
        api_url = f"https://api.ncbi.nlm.nih.gov/datasets/v1/taxonomy/taxon/{str(self._taxon)}"
        raw_response = requests.get(api_url, timeout=5)
        self._response = raw_response.json()["taxonomy_nodes"]

    def _collect_all_metadata(self):
        for child_metadata in self._response:
            taxonomy = child_metadata["taxonomy"]
            rank = taxonomy["rank"]
            name = taxonomy["organism_name"]
            tax_id = str(taxonomy["tax_id"])
            lineage = taxonomy["lineage"]
            if "Candidatus" not in name:
                if " sp. " not in name:
                    metadata = {
                        "sci_name": name,
                        "tax_id": tax_id,
                        "rank": rank,
                        "lineage": lineage,
                    }
                    self._all_metadata[tax_id] = metadata
                else:
                    logger.debug("{name} was not used for training", name=name)
            else:
                logger.debug("{name} was not used for training", name=name)

    def get_response(self):
        """Returns the raw response from the NCBI taxonomy database."""
        return self._response

    def get_metadata(self):
        """Returns the metadata from the NCBI taxonomy database."""
        return self._all_metadata



"""This module contains methods to download assemblies from the NCBI database."""

__author__ = "Berger, Phillip"

# pylint: disable=line-too-long

import os
import requests
from xspect.definitions import get_xspect_tmp_path


def download_assemblies(accessions, dir_name, target_folder, zip_file_name):
    """Download and save assemblies from the NCBI database.

    :param accessions: All collected accessions from the NCBI RefSeq-database.
    :type accessions: list
    :param dir_name: Name of the directory where the assemblies will be saved.
    :type dir_name: str
    :param target_folder: Name for the folder in which the downloaded files will be stored.
    :type target_folder: str
    :param zip_file_name: Name of the zip file. E.g. Klebsiella aerogenes.zip.
    :type zip_file_name: str
    """

    path = get_xspect_tmp_path() / dir_name / target_folder / zip_file_name
    api_url = f"https://api.ncbi.nlm.nih.gov/datasets/v1/genome/accession/{','.join(accessions)}/download"
    parameters = {"include_annotation_type": "GENOME_FASTA", "filename": zip_file_name}
    os.makedirs(os.path.dirname(path), exist_ok=True)
    genome_download = requests.get(api_url, params=parameters, timeout=20)
    with open(path, "wb") as f:
        f.write(genome_download.content)



"""Probabilistic filter model for sequence data"""

import json
from math import ceil
from pathlib import Path
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
from slugify import slugify
import cobs_index as cobs
from xspect.file_io import get_record_iterator
from xspect.models.result import ModelResult


class ProbabilisticFilterModel:
    """Probabilistic filter model for sequence data"""

    def __init__(
        self,
        k: int,
        model_display_name: str,
        author: str,
        author_email: str,
        model_type: str,
        base_path: Path,
        fpr: float = 0.01,
        num_hashes: int = 7,
    ) -> None:
        if k < 1:
            raise ValueError("Invalid k value, must be greater than 0")
        if not model_display_name:
            raise ValueError("Invalid filter display name, must be a non-empty string")
        if not model_type:
            raise ValueError("Invalid filter type, must be a non-empty string")
        if not isinstance(base_path, Path):
            raise ValueError("Invalid base path, must be a pathlib.Path object")

        self.k = k
        self.model_display_name = model_display_name
        self.author = author
        self.author_email = author_email
        self.model_type = model_type
        self.base_path = base_path
        self.display_names = {}
        self.fpr = fpr
        self.num_hashes = num_hashes
        self.index = None

    def get_cobs_index_path(self) -> Path:
        """Returns the path to the cobs index"""
        return str(self.base_path / self.slug() / "index.cobs_classic")

    def to_dict(self) -> dict:
        """Returns a dictionary representation of the model"""
        return {
            "k": self.k,
            "model_display_name": self.model_display_name,
            "author": self.author,
            "author_email": self.author_email,
            "model_type": self.model_type,
            "model_class": self.__class__.__name__,
            "display_names": self.display_names,
            "fpr": self.fpr,
            "num_hashes": self.num_hashes,
        }

    def __dict__(self) -> dict:
        """Returns a dictionary representation of the model"""
        return self.to_dict()

    def slug(self) -> str:
        """Returns a slug representation of the model"""
        return slugify(self.model_display_name + "-" + str(self.model_type))

    def fit(self, dir_path: Path, display_names: dict = None) -> None:
        """Adds filters to the model"""

        if display_names is None:
            display_names = {}

        if not isinstance(dir_path, Path):
            raise ValueError("Invalid directory path, must be a pathlib.Path object")

        if not dir_path.exists():
            raise ValueError("Directory path does not exist")

        if not dir_path.is_dir():
            raise ValueError("Directory path must be a directory")

        doclist = cobs.DocumentList()
        for file in dir_path.iterdir():
            if file.is_file() and file.suffix in [
                ".fasta",
                ".fna",
                ".fa",
                ".fastq",
                ".fq",
            ]:
                # cobs only uses the file name to the first "." as the document name
                if file.name in display_names:
                    self.display_names[file.name.split(".")[0]] = display_names[
                        file.name
                    ]
                else:
                    self.display_names[file.name.split(".")[0]] = file.stem
                doclist.add(str(file))

        if len(doclist) == 0:
            raise ValueError(
                "No valid files found in directory. Must be fasta or fastq"
            )

        index_params = cobs.ClassicIndexParameters()
        index_params.term_size = self.k
        index_params.num_hashes = self.num_hashes
        index_params.false_positive_rate = self.fpr
        index_params.clobber = True

        cobs.classic_construct_list(doclist, self.get_cobs_index_path(), index_params)

        self.index = cobs.Search(self.get_cobs_index_path(), True)

    def calculate_hits(
        self, sequence: Seq, filter_ids: list[str] = None, step: int = 1
    ) -> dict:
        """Calculates the hits for a sequence"""

        if not isinstance(sequence, (Seq)):
            raise ValueError(
                "Invalid sequence, must be a Bio.Seq or a Bio.SeqRecord object"
            )

        if not len(sequence) > self.k:
            raise ValueError("Invalid sequence, must be longer than k")

        r = self.index.search(str(sequence), step=step)
        result_dict = self._convert_cobs_result_to_dict(r)
        if filter_ids:
            return {doc: result_dict[doc] for doc in filter_ids}
        return result_dict

    def predict(
        self,
        sequence_input: (
            SeqRecord
            | list[SeqRecord]
            | SeqIO.FastaIO.FastaIterator
            | SeqIO.QualityIO.FastqPhredIterator
            | Path
        ),
        filter_ids: list[str] = None,
        step: int = 1,
    ) -> ModelResult:
        """Returns scores for the sequence(s) based on the filters in the model"""
        if isinstance(sequence_input, (SeqRecord)):
            return ProbabilisticFilterModel.predict(
                self, [sequence_input], filter_ids, step=step
            )

        if self._is_sequence_list(sequence_input) | self._is_sequence_iterator(
            sequence_input
        ):
            hits = {}
            num_kmers = {}
            for individual_sequence in sequence_input:
                individual_hits = self.calculate_hits(
                    individual_sequence.seq, filter_ids, step=step
                )
                num_kmers[individual_sequence.id] = self._count_kmers(
                    individual_sequence, step=step
                )
                hits[individual_sequence.id] = individual_hits
            return ModelResult(self.slug(), hits, num_kmers, sparse_sampling_step=step)

        if isinstance(sequence_input, Path):
            return ProbabilisticFilterModel.predict(
                self, get_record_iterator(sequence_input), step=step
            )

        raise ValueError(
            "Invalid sequence input, must be a Seq object, a list of Seq objects, a"
            " SeqIO FastaIterator, a SeqIO FastqPhredIterator, or a Path object to a"
            " fasta/fastq file"
        )

    def save(self) -> None:
        """Saves the model to disk"""
        json_path = self.base_path / f"{self.slug()}.json"
        filter_path = self.base_path / self.slug()
        filter_path.mkdir(exist_ok=True, parents=True)

        json_object = json.dumps(self.to_dict(), indent=4)

        with open(json_path, "w", encoding="utf-8") as file:
            file.write(json_object)

    @staticmethod
    def load(path: Path) -> "ProbabilisticFilterModel":
        """Loads the model from a file"""
        with open(path, "r", encoding="utf-8") as file:
            json_object = file.read()
            model_json = json.loads(json_object)
            model = ProbabilisticFilterModel(
                model_json["k"],
                model_json["model_display_name"],
                model_json["author"],
                model_json["author_email"],
                model_json["model_type"],
                path.parent,
                model_json["fpr"],
                model_json["num_hashes"],
            )
            model.display_names = model_json["display_names"]

            p = model.get_cobs_index_path()
            if not Path(p).exists():
                raise FileNotFoundError(f"Index file not found at {p}")
            model.index = cobs.Search(p, True)

            return model

    def _convert_cobs_result_to_dict(self, cobs_result: cobs.SearchResult) -> dict:
        return {
            individual_result.doc_name: individual_result.score
            for individual_result in cobs_result
        }

    def _count_kmers(
        self,
        sequence_input: (
            Seq
            | SeqRecord
            | list[Seq]
            | SeqIO.FastaIO.FastaIterator
            | SeqIO.QualityIO.FastqPhredIterator
        ),
        step: int = 1,
    ) -> int:
        """Counts the number of kmers in the sequence(s)"""
        if isinstance(sequence_input, Seq):
            return self._count_kmers([sequence_input], step=step)

        if isinstance(sequence_input, SeqRecord):
            return self._count_kmers(sequence_input.seq, step=step)

        is_sequence_list = isinstance(sequence_input, list) and all(
            isinstance(seq, Seq) for seq in sequence_input
        )
        is_iterator = isinstance(
            sequence_input,
            (SeqIO.FastaIO.FastaIterator, SeqIO.QualityIO.FastqPhredIterator),
        )

        if is_sequence_list | is_iterator:
            kmer_sum = 0
            for individual_sequence in sequence_input:
                # we need to look specifically at .seq for SeqIO iterators
                seq = individual_sequence.seq if is_iterator else individual_sequence
                num_kmers = ceil((len(seq) - self.k + 1) / step)
                kmer_sum += num_kmers
            return kmer_sum

        raise ValueError(
            "Invalid sequence input, must be a Seq object, a list of Seq objects, a"
            " SeqIO FastaIterator, or a SeqIO FastqPhredIterator"
        )

    def _is_sequence_list(self, sequence_input):
        return isinstance(sequence_input, list) and all(
            isinstance(seq, (SeqRecord)) for seq in sequence_input
        )

    def _is_sequence_iterator(self, sequence_input):
        return isinstance(
            sequence_input,
            (SeqIO.FastaIO.FastaIterator, SeqIO.QualityIO.FastqPhredIterator),
        )



"""Download filters from public repository."""

import os
import shutil
import requests

from xspect.definitions import get_xspect_model_path, get_xspect_tmp_path


def download_test_filters(url):
    """Download filters."""

    download_path = get_xspect_tmp_path() / "models.zip"
    extract_path = get_xspect_tmp_path() / "extracted_models"

    r = requests.get(url, allow_redirects=True, timeout=10)
    with open(download_path, "wb") as f:
        f.write(r.content)

    shutil.unpack_archive(
        download_path,
        extract_path,
        "zip",
    )

    shutil.copytree(
        extract_path,
        get_xspect_model_path(),
        dirs_exist_ok=True,
    )

    os.remove(download_path)
    shutil.rmtree(extract_path)



from pathlib import Path
from xspect.pipeline import Pipeline, PipelineStep, ModelExecution
from xspect.models.result import StepType


def test_save_and_load(tmpdir):
    """Test saving and loading a pipeline."""
    path = Path(tmpdir) / "test_pipeline.json"
    pipeline = Pipeline("Sample Pipeline", "John Doe", "john.doe@example.com")

    model_execution2 = ModelExecution("test_model2")
    pipeline_step = PipelineStep(
        StepType.FILTERING, "test_label", 0.7, model_execution2
    )
    model_execution = ModelExecution("test_model")
    model_execution.add_pipeline_step(pipeline_step)

    pipeline.save(path)

    loaded_pipeline = Pipeline.from_file(path)

    assert loaded_pipeline.display_name == pipeline.display_name
    assert loaded_pipeline.author == pipeline.author
    assert loaded_pipeline.author_email == pipeline.author_email
    assert len(loaded_pipeline.model_executions) == len(pipeline.model_executions)
    assert loaded_pipeline.to_dict() == pipeline.to_dict()



""" Tests for the ProbabilisticFilterSVMModel class. """

# pylint: disable=redefined-outer-name

from linecache import getline
from pathlib import Path
from csv import DictReader
import pytest
from src.xspect.models.probabilistic_filter_svm_model import ProbabilisticFilterSVMModel


@pytest.fixture
def filter_model(tmpdir):
    """Fixture for the ProbabilisticFilterSVMModel class."""
    base_path = Path(tmpdir.mkdir("xspect_data"))
    return ProbabilisticFilterSVMModel(
        k=21,
        model_display_name="Test Filter",
        author="John Doe",
        author_email="john.doe@example.com",
        model_type="Species",
        base_path=base_path,
        kernel="linear",
        c=1.0,
    )


@pytest.fixture
def trained_filter_model(
    filter_model, multiple_assembly_dir_path, multiple_assembly_svm_path
):
    """Fixture for the ProbabilisticFilterSVMModel class with trained model."""
    filter_model.fit(Path(multiple_assembly_dir_path), Path(multiple_assembly_svm_path))
    return filter_model


def test_fit(filter_model, multiple_assembly_dir_path, multiple_assembly_svm_path):
    """Test the fit method of the ProbabilisticFilterSVMModel class."""
    filter_model.fit(Path(multiple_assembly_dir_path), Path(multiple_assembly_svm_path))

    # Check if the scores.csv file has been created
    scores_file = filter_model.base_path / "test-filter-species" / "scores.csv"
    assert scores_file.is_file()

    with open(scores_file, "r", encoding="utf-8") as f:
        reader = DictReader(f)
        for row in reader:
            label_id = row["label_id"]
            assert row[label_id] == "1.0"

            # scores outside of the diagonal should be less than 1
            for key, value in row.items():
                if key not in ["label_id", "file", label_id]:
                    assert float(value) < 1


def test_predict(trained_filter_model, multiple_assembly_svm_path):
    """Test the predict method of the ProbabilisticFilterSVMModel class."""
    for file in Path(multiple_assembly_svm_path).iterdir():
        if file.suffix not in [".fasta", ".fa", ".fna", ".fastq", ".fq"]:
            continue
        res = trained_filter_model.predict(file, step=500)
        prediction = res.prediction
        file_header = getline(str(file), 1)
        label_id = file_header.replace("\n", "").replace(">", "")
        assert prediction == label_id


def test_set_svm_params(filter_model):
    """Test the set_svm_params method of the ProbabilisticFilterSVMModel class."""
    filter_model.set_svm_params(kernel="rbf", c=0.5)
    assert filter_model.kernel == "rbf"
    assert filter_model.c == 0.5


def test_save(filter_model):
    """Test the save method of the ProbabilisticFilterSVMModel class."""
    filter_model.save()
    json_path = filter_model.base_path / "test-filter-species.json"
    assert (json_path).is_file()
    with open(json_path, "r", encoding="utf-8") as f:
        data = f.read()
        assert "linear" in data
        assert "1.0" in data


def test_save_and_load(trained_filter_model):
    """Test the load method of the ProbabilisticFilterSVMModel class."""
    trained_filter_model.save()
    json_path = trained_filter_model.base_path / "test-filter-species.json"
    assert (json_path).is_file()

    loaded_model = ProbabilisticFilterSVMModel.load(json_path)
    assert trained_filter_model.to_dict() == loaded_model.to_dict()



""" Module for extracting and concatenating assemblies. """

__author__ = "Berger, Phillip"

import os
import shutil
from pathlib import Path
from Bio import SeqIO
from xspect import file_io
from xspect.definitions import get_xspect_tmp_path, fasta_endings


def change_header(assemblies_path, species_accessions: dict):
    """Change the header of the assemblies to the species name."""
    files = os.listdir(assemblies_path)
    # Iterate through all species.
    for name, accessions in species_accessions.items():
        # Iterate through all accessions of the current species.
        for accession in accessions:
            # Iterate through all file names.
            for file in files:
                if accession in file:
                    file_path = assemblies_path / str(file)
                    # Change the header.
                    with open(file_path, "r", encoding="utf-8") as f:
                        sequence = ""
                        for line in f.readlines():
                            if line[0] != ">":
                                sequence += line
                        new_header = f">{name}\n"
                    with open(file_path, "w", encoding="utf-8") as f:
                        f.write(new_header)
                        f.write(sequence)


def copy_assemblies(unzipped_path, assemblies_path):
    """Copy all assemblies to a new directory."""
    os.mkdir(assemblies_path)
    for folder in os.listdir(unzipped_path):
        for root, _, files in os.walk(unzipped_path / str(folder)):
            for file in files:
                file_ending = file.split(".")[-1]
                if file_ending in fasta_endings:
                    file_path = Path(root) / file
                    shutil.copy(file_path, (assemblies_path / file))


def concatenate_bf(unzipped_path, concatenate_path):
    """Concatenate all assemblies for Bloom filter training."""

    all_assemblies = []

    # Make new directory
    os.mkdir(concatenate_path)

    # Open the fasta files for each species.
    for folder in os.listdir(unzipped_path):
        species_files = []
        # Walk through dirs to get all fasta files.
        for root, _, files in os.walk(unzipped_path / folder):
            for file in files:
                file_ending = file.split(".")[-1]
                if file_ending in fasta_endings:
                    species_files.append(Path(root) / file)
                    all_assemblies.append(".".join(str(file).split(".")[:-1]))

        # Gather all sequences and headers.
        sequences = []
        headers = []
        for file in species_files:
            records = SeqIO.parse(file, "fasta")
            for record in records:
                headers.append(record.id)
                sequences.append(str(record.seq))

        # Concatenate sequences
        species_sequence = "".join(sequences)
        species_header = ">" + " § ".join(headers) + "\n"

        # Save concatenated sequences and headers
        species_path = concatenate_path / (folder + ".fasta")
        with open(species_path, "w", encoding="utf-8") as species_file:
            species_file.write(species_header)
            species_file.write(species_sequence)

    return all_assemblies


def save_all_assemblies(dir_path: Path, all_assemblies: list[str]):
    """Save all assemblies to a file."""
    path = dir_path / "all_bf_assemblies.txt"
    with open(path, "w", encoding="utf-8") as file:
        for assembly in all_assemblies:
            file.write(f"{assembly}\n")


def bf(dir_name: str, delete: bool):
    """Extract and concatenate assemblies for Bloom filter training."""
    dir_path = get_xspect_tmp_path() / dir_name
    zip_path = dir_path / "zip_files"
    unzipped_path = dir_path / "zip_files_extracted"
    concatenate_path = dir_path / "concatenate"
    file_io.extract_zip(zip_path, unzipped_path)
    all_assemblies = concatenate_bf(unzipped_path, concatenate_path)
    save_all_assemblies(dir_path, all_assemblies)
    if delete:
        file_io.delete_zip_files(zip_path)
        shutil.rmtree(zip_path, ignore_errors=False)
        shutil.rmtree(unzipped_path, ignore_errors=False)


def svm(species_accessions: dict, dir_name: str, delete: bool):
    """Extract and concatenate assemblies for generating SVM training data."""
    dir_path = get_xspect_tmp_path() / dir_name
    zip_path = dir_path / "training_data_zipped"
    unzipped_path = dir_path / "training_data_unzipped"
    assemblies_path = dir_path / "training_data"
    file_io.extract_zip(zip_path, unzipped_path)
    copy_assemblies(unzipped_path, assemblies_path)
    change_header(assemblies_path, species_accessions)
    if delete:
        file_io.delete_zip_files(zip_path)
        shutil.rmtree(zip_path, ignore_errors=False)
        shutil.rmtree(unzipped_path, ignore_errors=False)



""" HTML Scraping for the taxonomy check results from NCBI."""

__author__ = "Berger, Phillip"

import datetime
import pickle
import sys
import time
import requests
from loguru import logger
from xspect.definitions import get_xspect_root_path


class TaxonomyCheck:
    """Class to get the GCFs that passed the taxonomy check from NCBI."""

    _main_url = "https://ftp.ncbi.nlm.nih.gov/genomes/ASSEMBLY_REPORTS/ANI_report_prokaryotes.txt"
    _test_url = "https://ftp.ncbi.nlm.nih.gov/genomes/ASSEMBLY_REPORTS/"
    _main_path = get_xspect_root_path() / "taxonomy_check.txt"
    _test_path = get_xspect_root_path() / "tax_check_date.txt"
    _new_time: list
    _tax_check_ok = []

    def __init__(self):
        old_time = self._get_old_tax_date()
        self._new_time = self._get_new_tax_date()
        # Both Dates could be found.
        # Check if the html file was updated since the last time it was downloaded.
        # If yes than update the file.
        if self._new_time and old_time:
            if self._new_time == old_time:
                logger.info("File was not updated")
                self._get_old_file()
            else:
                logger.info("Updating file")
                self._update_tax_check()

        # The old date does not exist.
        # Get the html file for the taxonomy check results.
        elif self._new_time and not old_time:
            logger.info("No file was found. Creating new file")
            self._update_tax_check()

        elif not self._new_time and old_time:
            self._get_old_file()

        else:
            logger.error("Nothing was found")
            logger.error("Aborting")
            sys.exit()

    def _get_old_tax_date(self):
        try:
            with open(self._test_path, "rb") as f:
                old_time = pickle.load(f)
                return old_time
        except FileNotFoundError:
            return None

    def _get_new_tax_date(self):
        raw_response = requests.get(self._test_url, timeout=5)
        data = raw_response.text.split("\n")
        for line in data:
            if "ANI_report_prokaryotes.txt" in line:
                line_parts = line.split()
                date_parts = line_parts[-3].split("-")
                date = datetime.date(
                    int(date_parts[0]), int(date_parts[1]), int(date_parts[2])
                )
                time_parts = line_parts[-2].split(":")
                combined_time = datetime.time(int(time_parts[0]), int(time_parts[1]))
                new_time = [date, combined_time]

                return new_time

        return None

    def _update_tax_check(self):
        raw_response = requests.get(self._main_url, timeout=5)
        all_tax_checks = raw_response.text.split("\n")[1:-1]
        self._get_gcf_ok(all_tax_checks)
        self._save_time()
        self._save_file()

    def _get_gcf_ok(self, all_tax_checks: list):
        tax_check_ok = []
        for line in all_tax_checks:
            line_parts = line.split("\t")
            gcf = line_parts[1]
            tax_check_status = line_parts[-1]
            if tax_check_status == "OK":
                tax_check_ok.append(gcf)

        self._tax_check_ok = tax_check_ok

    def _save_time(self):
        with open(self._test_path, "wb") as f:
            pickle.dump(self._new_time, f)

    def _save_file(self):
        with open(self._main_path, "wb") as f:
            pickle.dump(self._tax_check_ok, f)

    def _get_old_file(self):
        with open(self._main_path, "rb") as f:
            self._tax_check_ok = pickle.load(f)

    @staticmethod
    def _get_current_time():
        return time.asctime(time.localtime()).split()[3]

    def ani_gcf(self):
        """Returns ANI GCFs that passed the taxonomy check."""
        return self._tax_check_ok



"""Probabilistic filter SVM model for sequence data"""

# pylint: disable=no-name-in-module, too-many-instance-attributes, arguments-renamed

import csv
import json
from linecache import getline
from pathlib import Path
from sklearn.svm import SVC
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
import cobs_index as cobs
from xspect.models.probabilistic_filter_model import ProbabilisticFilterModel
from xspect.definitions import fasta_endings, fastq_endings
from xspect.models.result import ModelResult


class ProbabilisticFilterSVMModel(ProbabilisticFilterModel):
    """Probabilistic filter SVM model for sequence data"""

    def __init__(
        self,
        k: int,
        model_display_name: str,
        author: str,
        author_email: str,
        model_type: str,
        base_path: Path,
        kernel: str,
        c: float,
        fpr: float = 0.01,
        num_hashes: int = 7,
    ) -> None:
        super().__init__(
            k=k,
            model_display_name=model_display_name,
            author=author,
            author_email=author_email,
            model_type=model_type,
            base_path=base_path,
            fpr=fpr,
            num_hashes=num_hashes,
        )
        self.kernel = kernel
        self.c = c

    def to_dict(self) -> dict:
        return super().to_dict() | {
            "kernel": self.kernel,
            "C": self.c,
        }

    def set_svm_params(self, kernel: str, c: float) -> None:
        """Set the parameters for the SVM"""
        self.kernel = kernel
        self.c = c
        self.save()

    def fit(
        self,
        dir_path: Path,
        svm_path: Path,
        display_names: dict = None,
        svm_step: int = 1,
    ) -> None:
        """Fit the SVM to the sequences and labels"""

        super().fit(dir_path, display_names=display_names)

        score_list = []
        for file in svm_path.iterdir():
            if not file.is_file():
                continue
            if file.suffix[1:] not in fasta_endings + fastq_endings:
                continue
            print(f"Calculating {file.name} scores for SVM training...")
            res = super().predict(file, step=svm_step)
            scores = res.get_scores()["total"]
            accession = "".join(file.name.split("_")[:2])
            file_header = getline(str(file), 1)
            label_id = file_header.replace("\n", "").replace(">", "")

            # format scores for csv
            scores = dict(sorted(scores.items()))
            scores = ",".join([str(score) for score in scores.values()])
            scores = f"{accession},{scores},{label_id}"
            score_list.append(scores)

        # csv header
        keys = list(self.display_names.keys())
        keys.sort()
        score_list.insert(0, f"file,{','.join(keys)},label_id")

        with open(
            self.base_path / self.slug() / "scores.csv", "w", encoding="utf-8"
        ) as file:
            file.write("\n".join(score_list))

    def predict(
        self,
        sequence_input: (
            SeqRecord
            | list[SeqRecord]
            | SeqIO.FastaIO.FastaIterator
            | SeqIO.QualityIO.FastqPhredIterator
            | Path
        ),
        filter_ids: list[str] = None,
        step: int = 1,
    ) -> dict:
        """Predict the labels of the sequences"""
        # get scores and format them for the SVM
        res = super().predict(sequence_input, filter_ids, step=step)
        svm_scores = dict(sorted(res.get_scores()["total"].items()))
        svm_scores = [list(svm_scores.values())]

        svm = self._get_svm(filter_ids)
        return ModelResult(
            self.slug(),
            res.hits,
            res.num_kmers,
            prediction=str(svm.predict(svm_scores)[0]),
        )

    def _get_svm(self, id_keys) -> SVC:
        """Get the SVM for the given id keys"""
        svm = SVC(kernel=self.kernel, C=self.c)
        # parse csv
        with open(
            self.base_path / self.slug() / "scores.csv", "r", encoding="utf-8"
        ) as file:
            file.readline()
            x_train = []
            y_train = []
            for row in csv.reader(file):
                if id_keys is None or row[-1] in id_keys:
                    x_train.append(row[1:-1])
                    y_train.append(row[-1])

        # train svm
        svm.fit(x_train, y_train)
        return svm

    @staticmethod
    def load(path: Path) -> "ProbabilisticFilterSVMModel":
        """Load the model from disk"""
        with open(path, "r", encoding="utf-8") as file:
            json_object = file.read()
            model_json = json.loads(json_object)
            model = ProbabilisticFilterSVMModel(
                model_json["k"],
                model_json["model_display_name"],
                model_json["author"],
                model_json["author_email"],
                model_json["model_type"],
                path.parent,
                model_json["kernel"],
                model_json["C"],
                fpr=model_json["fpr"],
                num_hashes=model_json["num_hashes"],
            )
            model.display_names = model_json["display_names"]

            p = model.get_cobs_index_path()
            if not Path(p).exists():
                raise FileNotFoundError(f"Index file not found at {p}")
            model.index = cobs.Search(p, True)

            return model



"""Tests for the ProbabilisticFilterModel class."""

# pylint: disable=redefined-outer-name, line-too-long, protected-access

from pathlib import Path
import pytest
from Bio.Seq import Seq
from src.xspect.models.probabilistic_single_filter_model import (
    ProbabilisticSingleFilterModel,
)


@pytest.fixture
def filter_model(tmpdir):
    """Fixture for the ProbabilisticFilterModel class."""
    base_path = tmpdir.mkdir("xspect_data")
    return ProbabilisticSingleFilterModel(
        k=21,
        model_display_name="Test Filter",
        author="John Doe",
        author_email="john.doe@example.com",
        model_type="Species",
        base_path=Path(base_path),
    )


@pytest.fixture
def trained_filter_model(filter_model, concatenated_assembly_file_path):
    """Fixture for the ProbabilisticFilterModel class."""
    filter_model.fit(Path(concatenated_assembly_file_path), "Acinetobacter baumannii")
    return filter_model


def test_fit(trained_filter_model):
    """Test the fit method."""
    assert trained_filter_model.bf is not None
    assert trained_filter_model.display_names == {
        "concatenated_assembly": "Acinetobacter baumannii"
    }


def test_calculate_hits(trained_filter_model):
    """Test the calculate_hits method."""
    hits = trained_filter_model.calculate_hits(Seq("TAAATAAATTTATATAGCTAAA"))
    assert hits == {"concatenated_assembly": 2}


def test_save_and_load(trained_filter_model):
    """Test the save and load methods."""
    trained_filter_model.save()
    loaded_filter_model = ProbabilisticSingleFilterModel.load(
        trained_filter_model.base_path / (trained_filter_model.slug() + ".json")
    )
    assert loaded_filter_model.to_dict() == trained_filter_model.to_dict()
    hits = trained_filter_model.calculate_hits(Seq("TAAATAAATTTATATAGCTAAA"))
    assert hits == {"concatenated_assembly": 2}



"""This module contains definitions for the XspecT package."""

from pathlib import Path
from os import getcwd

fasta_endings = ["fasta", "fna", "fa", "ffn", "frn"]
fastq_endings = ["fastq", "fq"]


def get_xspect_root_path():
    """Return the root path for XspecT data."""
    root_path = Path(getcwd()) / "xspect-data"
    root_path.mkdir(exist_ok=True, parents=True)
    return root_path


def get_xspect_model_path():
    """Return the path to the XspecT models."""
    model_path = get_xspect_root_path() / "models"
    model_path.mkdir(exist_ok=True, parents=True)
    return model_path


def get_xspect_tmp_path():
    """Return the path to the XspecT temporary files."""
    tmp_path = get_xspect_root_path() / "tmp"
    tmp_path.mkdir(exist_ok=True, parents=True)
    return tmp_path


def get_xspect_upload_path():
    """Return the path to the XspecT upload directory."""
    upload_path = get_xspect_root_path() / "uploads"
    upload_path.mkdir(exist_ok=True, parents=True)
    return upload_path


def get_xspect_runs_path():
    """Return the path to the XspecT runs directory."""
    runs_path = get_xspect_root_path() / "runs"
    runs_path.mkdir(exist_ok=True, parents=True)
    return runs_path



"""Project CLI"""

from pathlib import Path
import datetime
import click
import uvicorn
from xspect import fastapi
from xspect.download_filters import download_test_filters
from xspect.train import train_ncbi
from xspect.models.result import (
    StepType,
)
from xspect.definitions import get_xspect_runs_path, fasta_endings, fastq_endings
from xspect.pipeline import ModelExecution, Pipeline, PipelineStep


@click.group()
@click.version_option()
def cli():
    """XspecT CLI."""


@cli.command()
def download_filters():
    """Download filters."""
    click.echo("Downloading filters, this may take a while...")
    download_test_filters("https://xspect2.s3.eu-central-1.amazonaws.com/models.zip")


@cli.command()
@click.argument("genus")
@click.argument("path", type=click.Path(exists=True, dir_okay=True, file_okay=True))
@click.option(
    "-m",
    "--meta/--no-meta",
    help="Metagenome classification.",
    default=False,
)
@click.option(
    "-s",
    "--step",
    help="Sparse sampling step size (e. g. only every 500th kmer for step=500).",
    default=1,
)
def classify(genus, path, meta, step):
    """Classify sample(s) from file or directory PATH."""
    click.echo("Classifying...")
    click.echo(f"Step: {step}")

    file_paths = []
    if Path(path).is_dir():
        file_paths = [
            f
            for f in Path(path).iterdir()
            if f.is_file() and f.suffix[1:] in fasta_endings + fastq_endings
        ]
    else:
        file_paths = [Path(path)]

    # define pipeline
    pipeline = Pipeline(genus + " classification", "Test Author", "test@example.com")
    species_execution = ModelExecution(genus + "-species", sparse_sampling_step=step)
    if meta:
        species_filtering_step = PipelineStep(
            StepType.FILTERING, genus, 0.7, species_execution
        )
        genus_execution = ModelExecution(genus + "-genus", sparse_sampling_step=step)
        genus_execution.add_pipeline_step(species_filtering_step)
        pipeline.add_pipeline_step(genus_execution)
    else:
        pipeline.add_pipeline_step(species_execution)

    for idx, file_path in enumerate(file_paths):
        run = pipeline.run(file_path)
        time_str = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        save_path = get_xspect_runs_path() / f"run_{time_str}.json"
        run.save(save_path)
        print(
            f"[{idx+1}/{len(file_paths)}] Run finished. Results saved to '{save_path}'."
        )


@cli.command()
@click.argument("genus")
@click.option(
    "-bf-path",
    "--bf-assembly-path",
    help="Path to assembly directory for Bloom filter training.",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
)
@click.option(
    "-svm-path",
    "--svm-assembly-path",
    help="Path to assembly directory for SVM training.",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
)
@click.option(
    "-s",
    "--svm-step",
    help="SVM Sparse sampling step size (e. g. only every 500th kmer for step=500).",
    default=1,
)
def train(genus, bf_assembly_path, svm_assembly_path, svm_step):
    """Train model."""

    if bf_assembly_path or svm_assembly_path:
        raise NotImplementedError(
            "Training with specific assembly paths is not yet implemented."
        )
    try:
        train_ncbi(genus, svm_step=svm_step)
    except ValueError as e:
        raise click.ClickException(str(e)) from e


@cli.command()
def api():
    """Open the XspecT FastAPI."""
    uvicorn.run(fastapi.app, host="0.0.0.0", port=8000)


if __name__ == "__main__":
    cli()
